function [x,hist,t,NoMV]=gcr(A,b,x0,tol,kmax)
%x=gcr(A,b)
%  x is the computed solution (by gcr),
%  A is a matrix A or a function-handle, say A=@MV, 
%    that defines the action of A on a vector, with c=MV(u) we have that c=A*u;
%  b is the right-hand side vector.
%
%x=gcr(A,b,x0,tol,kmax)
%  x0 initial guess (0*b if not defined),
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b),
%  kmax maximal number of iterations (100 if not defined).
%
%[x,hist,t,NoMV]=gcr(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.
%

  %%% set defaults
  if nargin<5, kmax=100; end
  if nargin<4, tol=1.0e-8; end
  if nargin<3, x0=0*b; end
  
  %%% Is A is not a function-handle, then A is a matrix
  A_IS_MATRIX=(~isa(A,'function_handle')); NoMV=0;
  
  %%% define the matrix-vector multiplication MV
  function c=MV(x)
    if A_IS_MATRIX, c=A*x; else, c=feval(A,x); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end

     t0=clock; 
  %%%
  %%% Find dimension of the problem
  [n,m]=size(b); 
  %%% Allocate memory space
  c=zeros(n,kmax); u=zeros(n,kmax); sigma=zeros(1,kmax);

  x=x0; r=b; rho=norm(r);
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  %%% Update r0=r if x0 is non-zero
  if norm(x,1)>0, r=r-MV(x); rho=norm(r); end
      hist=rho;

  %%% The GCR loop
  for k=1:kmax+1
    if rho<tol, break, end
    u(:,k)=r;
    c(:,k)=MV(u(:,k));
    for j=1:k-1
      beta=(c(:,j)'*c(:,k))/sigma(j);
      u(:,k)=u(:,k)-beta*u(:,j);
      c(:,k)=c(:,k)-beta*c(:,j);
    end
    sigma(k)=c(:,k)'*c(:,k); alpha=(c(:,k)'*r)/sigma(k); 
    x=x+alpha*u(:,k);
    r=r-alpha*c(:,k);
       rho=norm(r); hist=[hist,rho];
  end
  
     t=etime(clock,t0);

end