%%% start with a clean memory
clear all
close all

%%% Define the problem A*x=b
%%%===========================================================
%%% To make sure that the quantities that define the matrix are 
%%% outside the loop of the iterative process of the solver, we 
%%% define a function that performs that matrix-vector multiplication (MV)
%%% (in, for instance, the file "problem1.m") and we use the handle
%%% (called A here in "main") to this MV in the call of the iterative solver.
%%%
%%% The files "problem1", "problem2", "problem3" define matrices that you can modify
%%% in your experiments (for this, set DEFINE_YOUR_OWN_MATRIX=1; below).
%%%
%%% The file "matrixhbo" generates predefined matrices that have been stored
%%% in the so-called Harwell-Boeing format (matrices in, e.g., the "Matrix Market"
%%% have also been stored in this format). Set DEFINE_YOUR_OWN_MATRIX=0 if
%%% you want to use these matrices in your experiments.
%%%
%%% You can also use matrices from MATLAB's gallery (run "help gallery" in MATLAB
%%% for more information. See also "problem0" and "problem0a").

DEFINE_YOUR_OWN_MATRIX=1;
if DEFINE_YOUR_OWN_MATRIX   %% Define your own matrix

  %%% set the dimension of the problem
  n=1000;
  
  %%% find the right-hand side vector b and 
  %%% the handle A of the function that performs the MV
  [A,b,Lambda]=problem1(n);

else %% Use a matrix defined in Harwell-Boeing format

   %%% select (the name of) the matrix
   NameMatrix='Grond2e3';  % NameMatrix='meier01';
   %%% find the handle A of the function that performs the MV
  [A,b,hbtype,Lambda]=matrixhbo(NameMatrix); b=b(:,1);
  fprintf(1,'\n\n%s: %d by %d matrix of type %s\n\n',NameMatrix,n,n,hbtype)
  
end

n=size(b,1);

%%%-------------------------------------------------------------

%%% If desired (for testing purposes), redefine the righthand side vector b
if 1
  %%% For ease of error checking, define b as A times the solution x
  % x=rand(n,1);
  x=ones(n,1);
  b=feval(A,x); b=b/norm(b);
end
%%%================================================================


%%% Is the interval Lambda that contains all eigenvalues available?
INTERVAL=exist('Lambda');
if INTERVAL, INTERVAL=(INTERVAL & length(Lambda)>1); end

if INTERVAL
   %%% set the degree of the Chebyshev polynomial
   ell=8;
   %%% find the zeros x of the Chebyshev polynomial of degree ell
   x=cos(((1:ell)-0.5)*pi/ell);
   %%% find interval, centre and radius of the interval that contains the eigenvalues 
   Lmin=min(Lambda); Lmax=max(Lambda);  %%% spectral ends
   mu=0.5*(Lmax+Lmin);                  %%% centre spectrum
   radius=0.5*(Lmax-Lmin);              %%% radius spectrum
   %%% shift the zeros x to interval Lambda that contains the eigenvalues 
   Mu=mu-radius*x;
end 

%%% str formats the output, displaying:
%%% the solution method, the dimension of the problem,
%%% required number of Matrix-vector multiplications, required solution time, 
%%% norm of the true residual, i.e., b-A*x with x the computed solution
%%% (norm of the residual as computed by the method)
str='%s\ndimension problem: %d\n#MVs: %d\ntime: %g\ntrue residual accuracy: %8.3g (%8.3g)\n\n';

%%% Solve using Polynomial solvers
%%% 1) Local Minimal Residuals
%%% Call solver
[x,hist,t,NoMV]=polynomialsolver(A,b,0*b,1.0e-8,1000);
%%% display convergence history
plot(0:length(hist)-1,log10(hist),'r-'), hold on
fprintf(str,'LMR ',n,NoMV,t,norm(b-feval(A,x),2),hist(end));

if INTERVAL
  %%% 2) Richardson with "optimal shift"
  [x,hist,t,NoMV]=polynomialsolver(A,b,0*b,1.0e-8,1000,mu);
  plot(0:length(hist)-1,log10(hist))
  fprintf(str,'Rich',n,NoMV,t,norm(b-feval(A,x),2),hist(end));
  %%% 3) Chebyshev
  [x,hist,t,NoMV]=polynomialsolver(A,b,0*b,1.0e-8,1000,Mu);
  plot(0:length(hist)-1,log10(hist),'m-')
  method=sprintf('Cheb(%d)',length(Mu));
  fprintf(str,method',n,NoMV,t,norm(b-feval(A,x),2),hist(end));
end

%%% Solve using a minimal residual solver
%%% GCR
[x,hist,t,NoMV]=gcr(A,b,0*b,1.0e-8,100);
plot(0:length(hist)-1,log10(hist),'k-*')
fprintf(str,'GCR ',n,NoMV,t,norm(b-feval(A,x),2),hist(end));


if INTERVAL
  legend('LMR','Rich',sprintf('Cheb(%d)',ell),'GCR')
else
  legend('LMR','GCR')
end

title(sprintf('Convergence history linear solvers. A is %d x %d',n,n))
xlabel('# MV')
ylabel('log_{10}||r_k||_2')


